# frozen_string_literal: true

require 'parser/ruby18'
require 'parser/ruby19'
require 'parser/ruby20'
require 'parser/ruby21'
require 'parser/ruby22'
require 'parser/ruby23'
require 'parser/ruby24'
require 'parser/ruby25'
require 'parser/ruby26'
require 'parser/ruby27'
require 'parser/ruby30'
require 'parser/ruby31'
require 'parser/ruby32'
